@@ -8,13 +8,14 @@
 #include <string.h>
 #include <unistd.h>
 #include <alloca.h>
+#include <inttypes.h>
 
 #include "do_mounts.h"
 #include "kinit.h"
+#include "fstype.h"
+#include "zlib.h"
 
-#define BUF_SZ		4096
-
-static const int do_devfs; // FIXME
+#define BUF_SZ		65536
 
 /* Find dev_t for e.g. "hda,NULL" or "hdb,2" */
 static dev_t
@@ -23,6 +24,7 @@ try_name(char *name, int part)
 	char path[BUF_SZ];
 	char buf[BUF_SZ];
 	int range;
+	unsigned int major, minor;
 	dev_t res;
 	char *s;
 	int len;
@@ -39,9 +41,13 @@ try_name(char *name, int part)
 	if (len <= 0 || len == BUF_SZ || buf[len - 1] != '\n')
 		goto fail;
 	buf[len - 1] = '\0';
-	res = (dev_t) strtoul(buf, &s, 16);
+	major = strtoul(buf, &s, 10);
+	if (*s != ':')
+		goto fail;
+	minor = strtoul(s+1, &s, 10);
 	if (*s)
 		goto fail;
+	res = makedev(major, minor);
 
 	/* if it's there and we are not looking for a partition - that's it */
 	if (!part)
@@ -62,8 +68,10 @@ try_name(char *name, int part)
 		goto fail;
 
 	/* if partition is within range - we got it */
-	if (part < range)
+	if (part < range) {
+		DEBUG(("kinit: try_name %s,%d = %#x\n", name, part, res+part));
 		return res + part;
+	}
 
 fail:
 	return (dev_t) 0;
@@ -89,8 +97,8 @@ fail:
  *	is mounted on rootfs /sys.
  */
 
-static dev_t
-name_to_dev_t(const char *name)
+static inline dev_t
+name_to_dev_t_real(const char *name)
 {
 	char *p;
 	dev_t res = 0;
@@ -98,11 +106,20 @@ name_to_dev_t(const char *name)
 	int part;
 	struct stat st;
 	int len;
+	const char *devname;
 
-	if ( name[0] == '/' && !stat(name, &st) && S_ISBLK(st.st_mode) )
+	if ( name[0] == '/' )
+		devname = name;
+	else {
+		char *dname = alloca(strlen(name) + 6);
+		sprintf(dname, "/dev/%s", name);
+		devname = dname;
+	}
+
+	if (!stat(devname, &st) && S_ISBLK(st.st_mode))
 		return st.st_rdev;
-	
-	if ( strncmp(name, "/dev/", 5) ) {
+
+	if (strncmp(name, "/dev/", 5)) {
 		res = (dev_t) strtoul(name, &p, 16);
 		if (*p)
 			return 0;
@@ -143,81 +160,21 @@ name_to_dev_t(const char *name)
 	return (dev_t)0;
 }
 
-static int
-find_in_devfs(char *path, dev_t dev)
+dev_t
+name_to_dev_t(const char *name)
 {
-	(void) path;
-	(void) dev;
-	
-#ifdef CONFIG_DEVFS_FS
-	// Leftover crap from the kernel.  Ugh.
+	dev_t dev = name_to_dev_t_real(name);
 
-	struct stat buf;
-	char *end = path + strlen(path);
-	int rest = path + 64 - end;
-	int size;
-	char *p = read_dir(path, &size);
-	char *s;
-
-	if (!p)
-		return -1;
-	for (s = p; s < p + size; s += ((struct linux_dirent64 *)s)->d_reclen) {
-		struct linux_dirent64 *d = (struct linux_dirent64 *)s;
-		if (strlen(d->d_name) + 2 > rest)
-			continue;
-		switch (d->d_type) {
-			case DT_BLK:
-				sprintf(end, "/%s", d->d_name);
-				if (sys_newstat(path, &buf) < 0)
-					break;
-				if (!S_ISBLK(buf.st_mode))
-					break;
-				if (buf.st_rdev != dev)
-					break;
-				kfree(p);
-				return 0;
-			case DT_DIR:
-				if (strcmp(d->d_name, ".") == 0)
-					break;
-				if (strcmp(d->d_name, "..") == 0)
-					break;
-				sprintf(end, "/%s", d->d_name);
-				if (find_in_devfs(path, dev) < 0)
-					break;
-				kfree(p);
-				return 0;
-		}
-	}
-	kfree(p);
-#endif
-	return -1;
+	DEBUG(("kinit: name_to_dev_t(%s) = %#x\n", name, dev));
+	return dev;
 }
 
 /* Create the device node "name" */
 int
-create_dev(const char *name, dev_t dev, const char *devfs_name)
+create_dev(const char *name, dev_t dev)
 {
-	char path[BUF_SZ];
-
 	unlink(name);
-	if (!do_devfs)
-		return mknod(name, S_IFBLK|0600, dev);
-
-	if (devfs_name && devfs_name[0]) {
-		if (strncmp(devfs_name, "/dev/", 5) == 0)
-			devfs_name += 5;
-		snprintf(path, BUF_SZ, "/dev/%s", devfs_name);
-		if (access(path, 0) == 0)
-			return symlink(devfs_name, name);
-	}
-	if (!dev)
-		return -1;
-
-	strcpy(path, "/dev");
-	if (find_in_devfs(path, dev) < 0)
-		return -1;
-
-	return symlink(path + 5, name);
+	return mknod(name, S_IFBLK|0600, dev);
 }
 
 /* mount a filesystem, possibly trying a set of different types */
@@ -231,6 +188,8 @@ mount_block(const char *source, const char *target,
 	ssize_t fsbytes;
 
 	if ( type ) {
+		DEBUG(("kinit: trying to mount %s on %s with type %s\n",
+		       source, target, type));
 		int rv = mount(source, target, type, flags, data);
 		/* Mount readonly if necessary */
 		if ( rv == -1 && errno == EACCES && !(flags & MS_RDONLY) )
@@ -257,7 +216,7 @@ mount_block(const char *source, const char *target,
 		*p++ = '\0';
 		if (*type != '\t')/* We can't mount a block device as a "nodev" fs */
 			continue;
-		
+
 		type++;
 		rp = mount_block(source, target, type, flags, data);
 		if ( rp )
@@ -279,7 +238,7 @@ mount_block_root(int argc, char *argv[], dev_t root_dev,
 	const char *rp;
 
 	data = get_arg(argc, argv, "rootflags=");
-	create_dev("/dev/root", root_dev, root_dev_name);
+	create_dev("/dev/root", root_dev);
 
 	errno = 0;
 
@@ -290,7 +249,7 @@ mount_block_root(int argc, char *argv[], dev_t root_dev,
 		if ( errno != EINVAL )
 			goto bad;
 	}
-	
+
 	if ( !errno && (rp = mount_block("/dev/root", "/root", NULL, flags, data)) )
 		goto ok;
 
@@ -321,10 +280,8 @@ mount_root(int argc, char *argv[], dev_t root_dev, const char *root_dev_name)
 	unsigned long flags = MS_RDONLY|MS_VERBOSE;
 	int ret;
 
-	if (get_arg(argc, argv, "ro")) {
-		flags |= MS_RDONLY;
-	}
-	if (get_arg(argc, argv, "rw")) {
+	if (get_flag(argc, argv, "rw")) {
+		DEBUG(("kinit: rw flag specified\n"));
 		flags &= ~MS_RDONLY;
 	}
 
@@ -337,29 +294,26 @@ mount_root(int argc, char *argv[], dev_t root_dev, const char *root_dev_name)
 	if (ret == 0) {
 		chdir("/root");
 	}
-	
-	return ret;
-}
 
-void
-md_run_setup(void)
-{
-	/* Muck around with md device(s) if necessary */
+	return ret;
 }
 
 int do_mounts(int argc, char *argv[])
 {
 	const char *root_dev_name = get_arg(argc, argv, "root=");
 	const char *root_delay = get_arg(argc, argv, "rootdelay=");
+	const char *load_ramdisk = get_arg(argc, argv, "load_ramdisk=");
 	dev_t root_dev = 0;
 
+	DEBUG(("kinit: do_mounts\n"));
+
 	if (root_delay) {
 		int delay = atoi(root_delay);
 		fprintf(stderr, "Waiting %d s before mounting root device...\n", delay);
 		sleep(delay);
 	}
 
-	md_run_setup();
+	md_run(argc, argv);
 
 	if (root_dev_name) {
 		root_dev = name_to_dev_t(root_dev_name);
@@ -371,11 +325,17 @@ int do_mounts(int argc, char *argv[])
 		root_dev = Root_NFS;
 	}
 
-	if ( initrd_load(argc, argv, root_dev) )
-		return 0;
+	DEBUG(("kinit: root_dev = %#x\n", root_dev));
 
-	/* If it's a floppy(only?) and load_ramdisk is set, then load
-	   a non-initrd ramdisk */
+	if ( initrd_load(argc, argv, root_dev) ) {
+		DEBUG(("initrd loaded\n"));
+		return 0;
+	}
+
+	if ( load_ramdisk && atoi(load_ramdisk) ) {
+		if (ramdisk_load(argc, argv, root_dev))
+			root_dev = Root_RAM0;
+	}
 
 	return mount_root(argc, argv, root_dev, root_dev_name);
 }
